searchNodes=[{"ref":"Breakfast.html","title":"Breakfast","type":"module","doc":"Breakfast is a decoder-generator library that: Has a consistent and declarative method for specifying the shape of your data Cuts down on boilerplate decoding code Leans on typespecs to determine how to validate data Provides clear error messages for invalid values Can be configured to decode any type of data In other words: describe what your data looks like, and Breakfast will generate a decoder for it. Table of Contents Use Case Quick Start Using Your Types Using the Result Custom Configuration Required Fields and Default Values Embedded Cereals Current State Contributing Use Case When dealing with some raw data, you might want to: Decode the data into a struct Validate that the types are what you expect them to be Have a typespec for your decoded data etc In Elixir, you might write the following to accomplish this: defmodule User do @type t :: %__MODULE__{ id: integer(), email: String.t(), roles: [String.t()] } @enforce_keys [:id, :email, :roles] defstruct @enforce_keys def decode(params) do with id when is_integer(id) &lt;- params[&quot;id&quot;], email when is_binary(email) &lt;- params[&quot;email&quot;], roles when is_list(roles) &lt;- params[&quot;roles&quot;], true &lt;- Enum.all?(roles, &amp;is_binary/1) do {:ok, %__MODULE__{id: id, email: email, roles: roles}} else _ -&gt; :error end end end iex&gt; data = %{ ...&gt; &quot;id&quot; =&gt; 1, ...&gt; &quot;email&quot; =&gt; &quot;john@aol.com&quot;, ...&gt; &quot;roles&quot; =&gt; [&quot;admin&quot;, &quot;exec&quot;] ...&gt; } ...&gt; User.decode(data) {:ok, %User{id: 1, email: &quot;john@aol.com&quot;, roles: [&quot;admin&quot;, &quot;exec&quot;]}} iex&gt; data = %{ ...&gt; &quot;id&quot; =&gt; 1, ...&gt; &quot;email&quot; =&gt; &quot;john@aol.com&quot;, ...&gt; &quot;roles&quot; =&gt; [&quot;admin&quot;, :exec] ...&gt; } ...&gt; User.decode(data) :error With Breakfast, you can get the same (and more) just by describing what your data should look like: defmodule User do use Breakfast cereal do field :id, integer() field :email, String.t() field :roles, [String.t()] end end iex&gt; data = %{ ...&gt; &quot;id&quot; =&gt; 1, ...&gt; &quot;email&quot; =&gt; &quot;john@aol.com&quot;, ...&gt; &quot;roles&quot; =&gt; [&quot;admin&quot;, &quot;exec&quot;] ...&gt; } ...&gt; Breakfast.decode(User, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;email&quot; =&gt; &quot;john@aol.com&quot;, &quot;id&quot; =&gt; 1, &quot;roles&quot; =&gt; [&quot;admin&quot;, &quot;exec&quot;]}, struct: %User{email: &quot;john@aol.com&quot;, id: 1, roles: [&quot;admin&quot;, &quot;exec&quot;]} } iex&gt; data = %{ ...&gt; &quot;id&quot; =&gt; 1, ...&gt; &quot;email&quot; =&gt; &quot;john@aol.com&quot;, ...&gt; &quot;roles&quot; =&gt; [&quot;admin&quot;, :exec] ...&gt; } ...&gt; Breakfast.decode(User, data) %Breakfast.Yogurt{ errors: [roles: &quot;expected a list of type :binary, got a list with at least one invalid element: expected a binary, got: :exec&quot;], params: %{&quot;email&quot; =&gt; &quot;john@aol.com&quot;, &quot;id&quot; =&gt; 1, &quot;roles&quot; =&gt; [&quot;admin&quot;, :exec]}, struct: %User{email: &quot;john@aol.com&quot;, id: 1, roles: nil} } Quick Start Installing Before you do anything, you need to add :breakfast as a dependency in your mix.exs file: # mix.exs defp deps do [ {:breakfast, &quot;0.1.1&quot;} ] end Decoding with Breakfast Let&#39;s say you&#39;re trying to decode some data of the following shape: %{ &quot;email&quot; =&gt; &quot;leo@aol.com&quot;, &quot;age&quot; =&gt; 67, &quot;roles&quot; =&gt; [&quot;exec&quot;, &quot;admin&quot;] } First, we need to define a decoder that describes the shape of this data. Breakfast&#39;s interface for describing the shape of your data is very similar to Ecto&#39;s Schema definitions. The primary difference between Breakfast and Ecto schemas is that Breakfast leans on Elixir Typespecs to declare your data&#39;s types. Here is a simple example of describing the shape of the above data with Breakfast: defmodule User do use Breakfast cereal do field :email, String.t() field :age, non_neg_integer() field :roles, [String.t()] end end This decoder module is what Breakfast will use to decode and validate your data. Once it&#39;s defined, you can pass this module along with the raw params to Breakfast.decode/2: defmodule User do use Breakfast cereal do field :email, String.t() field :age, non_neg_integer() field :roles, [String.t()] end end iex&gt; data = %{ ...&gt; &quot;email&quot; =&gt; &quot;leo@aol.com&quot;, ...&gt; &quot;age&quot; =&gt; 67, ...&gt; &quot;roles&quot; =&gt; [&quot;exec&quot;, &quot;admin&quot;] ...&gt; } ...&gt; Breakfast.decode(User, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;age&quot; =&gt; 67, &quot;email&quot; =&gt; &quot;leo@aol.com&quot;, &quot;roles&quot; =&gt; [&quot;exec&quot;, &quot;admin&quot;]}, struct: %User{age: 67, email: &quot;leo@aol.com&quot;, roles: [&quot;exec&quot;, &quot;admin&quot;]} } That&#39;s it! Breakfast can decode basic data with little configuration, but can be told to do a lot more. Using Your Types Beyond documenting your data, the typespecs for each field are also used to automatically determine how to validate that field. In the following example, we can see that a field of type non_neg_integer() will not accept a value &lt; 0: defmodule User do use Breakfast cereal do field :name, String.t() field :age, non_neg_integer() end end iex&gt; data = %{ ...&gt; &quot;name&quot; =&gt; &quot;Sean&quot;, ...&gt; &quot;age&quot; =&gt; -5 ...&gt; } ...&gt; Breakfast.decode(User, data) %Breakfast.Yogurt{ errors: [age: &quot;expected a non_neg_integer, got: -5&quot;], params: %{&quot;age&quot; =&gt; -5, &quot;name&quot; =&gt; &quot;Sean&quot;}, struct: %User{age: nil, name: &quot;Sean&quot;} } Breakfast can even handle more complex types, such as unions: defmodule Request do use Breakfast cereal do field :payload, map() field :status, :pending | :success | :failed end end iex&gt; data = %{ ...&gt; &quot;payload&quot; =&gt; %{&quot;some&quot; =&gt; &quot;data&quot;}, ...&gt; &quot;status&quot; =&gt; :success ...&gt; } ...&gt; Breakfast.decode(Request, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;payload&quot; =&gt; %{&quot;some&quot; =&gt; &quot;data&quot;}, &quot;status&quot; =&gt; :success}, struct: %Request{payload: %{&quot;some&quot; =&gt; &quot;data&quot;}, status: :success} } iex&gt; data = %{ ...&gt; &quot;payload&quot; =&gt; %{&quot;some&quot; =&gt; &quot;data&quot;}, ...&gt; &quot;status&quot; =&gt; :waiting ...&gt; } ...&gt; Breakfast.decode(Request, data) %Breakfast.Yogurt{ errors: [status: &quot;expected one of :pending | :success | :failed, got: :waiting&quot;], params: %{&quot;payload&quot; =&gt; %{&quot;some&quot; =&gt; &quot;data&quot;}, &quot;status&quot; =&gt; :waiting}, struct: %Request{payload: %{&quot;some&quot; =&gt; &quot;data&quot;}, status: nil} } Checkout the types docs for more on what types Breakfast supports. Using the Result You might be asking, what&#39;s this %Yogurt{} thing? A %Yogurt{} represents the result of a decoding. It contains three pieces of data: params: The original input params that you asked Breakfast to decode errors: A list of human-readable string errors that were accumulated when trying to decode the params struct: The decoded data that&#39;s been casted to the well-defined struct In your day-to-day programming, you can pattern match on a %Yogurt{} for control-flow, where an empty :errors list indicates that the decoding was successful: defmodule MathRequest do use Breakfast cereal do field :lhs, number() field :rhs, number() field :operation, :+ | :- | :* | :/, cast: :existing_atom_from_string end def existing_atom_from_string(value) do {:ok, String.to_existing_atom(value)} rescue _ in ArgumentError -&gt; :error end end iex&gt; request = %{&quot;lhs&quot; =&gt; 5.0, &quot;rhs&quot; =&gt; 2, &quot;operation&quot; =&gt; &quot;/&quot;} ...&gt; case Breakfast.decode(MathRequest, request) do ...&gt; %Breakfast.Yogurt{errors: [], struct: result} -&gt; {:ok, result} ...&gt; %Breakfast.Yogurt{errors: errors} -&gt; {:error, errors} ...&gt; end {:ok, %MathRequest{lhs: 5.0, rhs: 2, operation: :/}} iex&gt; request = %{&quot;lhs&quot; =&gt; 5.0, &quot;rhs&quot; =&gt; 2, &quot;operation&quot; =&gt; &quot;%&quot;} ...&gt; case Breakfast.decode(MathRequest, request) do ...&gt; %Breakfast.Yogurt{errors: [], struct: result} -&gt; {:ok, result} ...&gt; %Breakfast.Yogurt{errors: errors} -&gt; {:error, errors} ...&gt; end {:error, [operation: &quot;expected one of :+ | :- | :* | :/, got: :%&quot;]} What about :ok | :error tuples? We decided to not use :ok | :error tuples as the return type for the following reasons: We wanted to have a consistent type for the return value (it&#39;s always a Yogurt.t(), no matter what) There&#39;s a lot of context to return that you may or may not want to use (i.e. errors, input params, etc) You can still pattern match on any case that you care about handling in your code Custom Configuration When Breakfast is decoding data, it runs through the same 3 steps for each field: fetch: Retrieve the field value from the data (i.e. Map.fetch/2) cast: Map the field value from one value to another, if necessary (i.e. Integer.parse/1) validate: Check to see if the field value is valid (i.e. is_binary/1) Out of the box, Breakfast will assume the following for each step: fetch: The raw data is in the form of a string-keyed map, and the key for the field is the string version of the declared field name cast: No casting is necessary validate: Ensure that the value matches the field type However, each of these steps can be customized for any field: defmodule Settings do use Breakfast cereal do field(:name, String.t(), fetch: :fetch_name) field(:timeout, integer(), cast: :int_from_string) field(:volume, integer(), validate: :valid_volume) end def fetch_name(params, :name) do Map.fetch(params, &quot;SettingsName&quot;) end def int_from_string(value) do with true &lt;- is_binary(value), {int, &quot;&quot;} &lt;- Integer.parse(value) do {:ok, int} else _ -&gt; :error end end def valid_volume(volume) when volume in 0..100, do: [] def valid_volume(volume), do: [&quot;expected an integer in 0..100, got: \#{inspect(volume)}&quot;] end iex&gt; data = %{ ...&gt; &quot;SettingsName&quot; =&gt; &quot;Control Pannel&quot;, ...&gt; &quot;timeout&quot; =&gt; &quot;1500&quot;, ...&gt; &quot;volume&quot; =&gt; 8 ...&gt; } ...&gt; Breakfast.decode(Settings, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;SettingsName&quot; =&gt; &quot;Control Pannel&quot;, &quot;timeout&quot; =&gt; &quot;1500&quot;, &quot;volume&quot; =&gt; 8}, struct: %Settings{name: &quot;Control Pannel&quot;, timeout: 1500, volume: 8} } iex&gt; data = %{ ...&gt; &quot;name&quot; =&gt; &quot;Control Pannel&quot;, ...&gt; &quot;timeout&quot; =&gt; 1500, ...&gt; &quot;volume&quot; =&gt; -100 ...&gt; } ...&gt; Breakfast.decode(Settings, data) %Breakfast.Yogurt{ errors: [name: &quot;value not found&quot;, timeout: &quot;cast error&quot;, volume: &quot;expected an integer in 0..100, got: -100&quot;], params: %{&quot;name&quot; =&gt; &quot;Control Pannel&quot;, &quot;timeout&quot; =&gt; 1500, &quot;volume&quot; =&gt; -100}, struct: %Settings{name: nil, timeout: nil, volume: nil} } You can also set the default behaviour for any of these steps: defmodule RGBColor do use Breakfast cereal fetch: :fetch_upcase_key, cast: :int_from_string, validate: :valid_rgb_value do field :r, integer() field :g, integer() field :b, integer() end def fetch_upcase_key(params, field) do key = field |&gt; to_string() |&gt; String.upcase() Map.fetch(params, key) end def int_from_string(value) do with true &lt;- is_binary(value), {int, &quot;&quot;} &lt;- Integer.parse(value) do {:ok, int} else _ -&gt; :error end end def valid_rgb_value(value) when value in 0..255, do: [] def valid_rgb_value(value), do: [&quot;expected an integer between 0 and 255, got: \#{inspect(value)}&quot;] end iex&gt; data = %{&quot;R&quot; =&gt; &quot;10&quot;, &quot;G&quot; =&gt; &quot;20&quot;, &quot;B&quot; =&gt; &quot;30&quot;} ...&gt; Breakfast.decode(RGBColor, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;B&quot; =&gt; &quot;30&quot;, &quot;G&quot; =&gt; &quot;20&quot;, &quot;R&quot; =&gt; &quot;10&quot;}, struct: %RGBColor{b: 30, g: 20, r: 10} } iex&gt; data = %{&quot;r&quot; =&gt; &quot;10&quot;, &quot;G&quot; =&gt; &quot;Twenty&quot;, &quot;B&quot; =&gt; &quot;500&quot;} ...&gt; Breakfast.decode(RGBColor, data) %Breakfast.Yogurt{ errors: [r: &quot;value not found&quot;, g: &quot;cast error&quot;, b: &quot;expected an integer between 0 and 255, got: 500&quot;], params: %{&quot;B&quot; =&gt; &quot;500&quot;, &quot;G&quot; =&gt; &quot;Twenty&quot;, &quot;r&quot; =&gt; &quot;10&quot;}, struct: %RGBColor{b: nil, g: nil, r: nil} } Given this, Breakfast can actually decode any form of data, not just maps: defmodule SpreadsheetRow do use Breakfast @column_indices %{ name: 0, age: 1, email: 2 } cereal fetch: :fetch_at_list_index do field :name, String.t() field :age, non_neg_integer() field :email, String.t() end def fetch_at_list_index(data, field_name) do index = Map.fetch!(@column_indices, field_name) Enum.fetch(data, index) end end iex&gt; data = [&quot;Sully&quot;, 37, &quot;sully@aol.com&quot;] ...&gt; Breakfast.decode(SpreadsheetRow, data) %Breakfast.Yogurt{ errors: [], params: [&quot;Sully&quot;, 37, &quot;sully@aol.com&quot;], struct: %SpreadsheetRow{age: 37, email: &quot;sully@aol.com&quot;, name: &quot;Sully&quot;} } Required Fields and Default Values By default, Breakfast considers every field to be a required field. The only way to make a field &quot;optional&quot; is to provide a :default value for that field: defmodule Post do use Breakfast cereal do field :title, String.t() field :content, String.t() field :tags, [String.t()], default: [] end end iex&gt; data = %{ ...&gt; &quot;title&quot; =&gt; &quot;Cool Thing I Did&quot;, ...&gt; &quot;content&quot; =&gt; &quot;Thanks for reading!&quot;, ...&gt; } ...&gt; Breakfast.decode(Post, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;content&quot; =&gt; &quot;Thanks for reading!&quot;, &quot;title&quot; =&gt; &quot;Cool Thing I Did&quot;}, struct: %Post{ content: &quot;Thanks for reading!&quot;, tags: [], title: &quot;Cool Thing I Did&quot; } } iex&gt; data = %{ ...&gt; &quot;title&quot; =&gt; &quot;Cool Thing I Did&quot;, ...&gt; &quot;content&quot; =&gt; &quot;Thanks for reading!&quot;, ...&gt; &quot;tags&quot; =&gt; [&quot;blockchain&quot;, &quot;crypto&quot;] ...&gt; } ...&gt; Breakfast.decode(Post, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;content&quot; =&gt; &quot;Thanks for reading!&quot;, &quot;tags&quot; =&gt; [&quot;blockchain&quot;, &quot;crypto&quot;], &quot;title&quot; =&gt; &quot;Cool Thing I Did&quot;}, struct: %Post{ content: &quot;Thanks for reading!&quot;, tags: [&quot;blockchain&quot;, &quot;crypto&quot;], title: &quot;Cool Thing I Did&quot; } } Embedded Cereals Breakfast allows you to use decoders within each other to describe the shape of nested data. Here, the Config decoder is used as the type for the User.config field: defmodule Player do defmodule Config do use Breakfast cereal do field :timezone, String.t() field :sleep_timeout, non_neg_integer() end end use Breakfast cereal do field :name, String.t() field :score, integer() field :config, {:cereal, Config} end end iex&gt; data = %{ ...&gt; &quot;name&quot; =&gt; &quot;Leo&quot;, ...&gt; &quot;score&quot; =&gt; 1600, ...&gt; &quot;config&quot; =&gt; %{ ...&gt; &quot;timezone&quot; =&gt; &quot;EST&quot;, ...&gt; &quot;sleep_timeout&quot; =&gt; 5000 ...&gt; } ...&gt; } ...&gt; Breakfast.decode(Player, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;config&quot; =&gt; %{&quot;sleep_timeout&quot; =&gt; 5000, &quot;timezone&quot; =&gt; &quot;EST&quot;}, &quot;name&quot; =&gt; &quot;Leo&quot;, &quot;score&quot; =&gt; 1600}, struct: %Player{ name: &quot;Leo&quot;, score: 1600, config: %Player.Config{ sleep_timeout: 5000, timezone: &quot;EST&quot; } } } iex&gt; data = %{ ...&gt; &quot;name&quot; =&gt; &quot;Leo&quot;, ...&gt; &quot;score&quot; =&gt; 1600, ...&gt; &quot;config&quot; =&gt; %{ ...&gt; &quot;timezone&quot; =&gt; &quot;EST&quot;, ...&gt; &quot;sleep_timeout&quot; =&gt; -5000 ...&gt; } ...&gt; } ...&gt; Breakfast.decode(Player, data) %Breakfast.Yogurt{ errors: [config: [sleep_timeout: &quot;expected a non_neg_integer, got: -5000&quot;]], params: %{&quot;config&quot; =&gt; %{&quot;sleep_timeout&quot; =&gt; -5000, &quot;timezone&quot; =&gt; &quot;EST&quot;}, &quot;name&quot; =&gt; &quot;Leo&quot;, &quot;score&quot; =&gt; 1600}, struct: %Player{config: nil, name: &quot;Leo&quot;, score: 1600} } Current State Breakfast 0.1 has been released! Further v0.1.x versions will include bug fixes, enhancements, etc. Breakfast 0.2 development will include all new major features and breaking changes. Check out out the roadmap to see what&#39;s coming/if you are looking to contribute! Contributing Contributions are extremely welcome! This can take the form of pull requests and/or opening issues for bugs, feature requests, or general discussion. If you want to make some changes but aren&#39;t sure where to begin, I&#39;d be happy to help :). I&#39;d like to thank the following people who contributed to this project either via code and/or good ideas: @evuez @GeoffreyPS"},{"ref":"Breakfast.html#decode/2","title":"Breakfast.decode/2","type":"function","doc":"The function to call when you want to decode some data. It requires that you&#39;ve defined your decoder module and takes the module name as a parameter. See the quick start guide for more on defining and user decoder modules. Examples iex&gt; defmodule Customer do ...&gt; use Breakfast ...&gt; ...&gt; cereal do ...&gt; field :id, non_neg_integer() ...&gt; field :email, String.t() ...&gt; end ...&gt; end ...&gt; ...&gt; data = %{ ...&gt; &quot;id&quot; =&gt; 5, ...&gt; &quot;email&quot; =&gt; &quot;leo@aol.com&quot; ...&gt; } ...&gt; ...&gt; Breakfast.decode(Customer, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;email&quot; =&gt; &quot;leo@aol.com&quot;, &quot;id&quot; =&gt; 5}, struct: %Customer{email: &quot;leo@aol.com&quot;, id: 5} }"},{"ref":"Breakfast.CastError.html","title":"Breakfast.CastError","type":"exception","doc":"Raised at runtime when a caster returns something other than {:ok, value} or :error."},{"ref":"Breakfast.CompileError.html","title":"Breakfast.CompileError","type":"exception","doc":"Raised at compilation time when the cereal cannot be compiled."},{"ref":"Breakfast.FetchError.html","title":"Breakfast.FetchError","type":"exception","doc":"Raised at runtime when a fetcher returns something other than {:ok, value} or :error."},{"ref":"Breakfast.TypeError.html","title":"Breakfast.TypeError","type":"exception","doc":"Raised at compilation time when a type cannot be derived."},{"ref":"Breakfast.ValidateError.html","title":"Breakfast.ValidateError","type":"exception","doc":"Raised at runtime when a validator returns something other than a list."},{"ref":"Breakfast.Yogurt.html","title":"Breakfast.Yogurt","type":"module","doc":"The %Elixir.Breakfast.Yogurt{} struct is the resulting value of a call to Breakfast.decode/2. It contains all the information regarding a particular decode, including: the params passed in errors that occured for a given field the resulting, decoded struct"},{"ref":"Breakfast.Yogurt.html#valid?/1","title":"Breakfast.Yogurt.valid?/1","type":"function","doc":""},{"ref":"Breakfast.Yogurt.html#t:t/0","title":"Breakfast.Yogurt.t/0","type":"type","doc":""},{"ref":"readme.html","title":"README","type":"extras","doc":"Breakfast is a decoder-generator library that: Has a consistent and declarative method for specifying the shape of your data Cuts down on boilerplate decoding code Leans on typespecs to determine how to validate data Provides clear error messages for invalid values Can be configured to decode any type of data In other words: describe what your data looks like, and Breakfast will generate a decoder for it."},{"ref":"readme.html#table-of-contents","title":"README - Table of Contents","type":"extras","doc":"Use Case Quick Start Using Your Types Using the Result Custom Configuration Required Fields and Default Values Embedded Cereals Current State Contributing"},{"ref":"readme.html#use-case","title":"README - Use Case","type":"extras","doc":"When dealing with some raw data, you might want to: Decode the data into a struct Validate that the types are what you expect them to be Have a typespec for your decoded data etc In Elixir, you might write the following to accomplish this: defmodule User do @type t :: %__MODULE__{ id: integer(), email: String.t(), roles: [String.t()] } @enforce_keys [:id, :email, :roles] defstruct @enforce_keys def decode(params) do with id when is_integer(id) &lt;- params[&quot;id&quot;], email when is_binary(email) &lt;- params[&quot;email&quot;], roles when is_list(roles) &lt;- params[&quot;roles&quot;], true &lt;- Enum.all?(roles, &amp;is_binary/1) do {:ok, %__MODULE__{id: id, email: email, roles: roles}} else _ -&gt; :error end end end iex&gt; data = %{ ...&gt; &quot;id&quot; =&gt; 1, ...&gt; &quot;email&quot; =&gt; &quot;john@aol.com&quot;, ...&gt; &quot;roles&quot; =&gt; [&quot;admin&quot;, &quot;exec&quot;] ...&gt; } ...&gt; User.decode(data) {:ok, %User{id: 1, email: &quot;john@aol.com&quot;, roles: [&quot;admin&quot;, &quot;exec&quot;]}} iex&gt; data = %{ ...&gt; &quot;id&quot; =&gt; 1, ...&gt; &quot;email&quot; =&gt; &quot;john@aol.com&quot;, ...&gt; &quot;roles&quot; =&gt; [&quot;admin&quot;, :exec] ...&gt; } ...&gt; User.decode(data) :error With Breakfast, you can get the same (and more) just by describing what your data should look like: defmodule User do use Breakfast cereal do field :id, integer() field :email, String.t() field :roles, [String.t()] end end iex&gt; data = %{ ...&gt; &quot;id&quot; =&gt; 1, ...&gt; &quot;email&quot; =&gt; &quot;john@aol.com&quot;, ...&gt; &quot;roles&quot; =&gt; [&quot;admin&quot;, &quot;exec&quot;] ...&gt; } ...&gt; Breakfast.decode(User, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;email&quot; =&gt; &quot;john@aol.com&quot;, &quot;id&quot; =&gt; 1, &quot;roles&quot; =&gt; [&quot;admin&quot;, &quot;exec&quot;]}, struct: %User{email: &quot;john@aol.com&quot;, id: 1, roles: [&quot;admin&quot;, &quot;exec&quot;]} } iex&gt; data = %{ ...&gt; &quot;id&quot; =&gt; 1, ...&gt; &quot;email&quot; =&gt; &quot;john@aol.com&quot;, ...&gt; &quot;roles&quot; =&gt; [&quot;admin&quot;, :exec] ...&gt; } ...&gt; Breakfast.decode(User, data) %Breakfast.Yogurt{ errors: [roles: &quot;expected a list of type :binary, got a list with at least one invalid element: expected a binary, got: :exec&quot;], params: %{&quot;email&quot; =&gt; &quot;john@aol.com&quot;, &quot;id&quot; =&gt; 1, &quot;roles&quot; =&gt; [&quot;admin&quot;, :exec]}, struct: %User{email: &quot;john@aol.com&quot;, id: 1, roles: nil} }"},{"ref":"readme.html#quick-start","title":"README - Quick Start","type":"extras","doc":"Installing Before you do anything, you need to add :breakfast as a dependency in your mix.exs file: # mix.exs defp deps do [ {:breakfast, &quot;0.1.1&quot;} ] end Decoding with Breakfast Let&#39;s say you&#39;re trying to decode some data of the following shape: %{ &quot;email&quot; =&gt; &quot;leo@aol.com&quot;, &quot;age&quot; =&gt; 67, &quot;roles&quot; =&gt; [&quot;exec&quot;, &quot;admin&quot;] } First, we need to define a decoder that describes the shape of this data. Breakfast&#39;s interface for describing the shape of your data is very similar to Ecto&#39;s Schema definitions. The primary difference between Breakfast and Ecto schemas is that Breakfast leans on Elixir Typespecs to declare your data&#39;s types. Here is a simple example of describing the shape of the above data with Breakfast: defmodule User do use Breakfast cereal do field :email, String.t() field :age, non_neg_integer() field :roles, [String.t()] end end This decoder module is what Breakfast will use to decode and validate your data. Once it&#39;s defined, you can pass this module along with the raw params to Breakfast.decode/2: defmodule User do use Breakfast cereal do field :email, String.t() field :age, non_neg_integer() field :roles, [String.t()] end end iex&gt; data = %{ ...&gt; &quot;email&quot; =&gt; &quot;leo@aol.com&quot;, ...&gt; &quot;age&quot; =&gt; 67, ...&gt; &quot;roles&quot; =&gt; [&quot;exec&quot;, &quot;admin&quot;] ...&gt; } ...&gt; Breakfast.decode(User, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;age&quot; =&gt; 67, &quot;email&quot; =&gt; &quot;leo@aol.com&quot;, &quot;roles&quot; =&gt; [&quot;exec&quot;, &quot;admin&quot;]}, struct: %User{age: 67, email: &quot;leo@aol.com&quot;, roles: [&quot;exec&quot;, &quot;admin&quot;]} } That&#39;s it! Breakfast can decode basic data with little configuration, but can be told to do a lot more."},{"ref":"readme.html#using-your-types","title":"README - Using Your Types","type":"extras","doc":"Beyond documenting your data, the typespecs for each field are also used to automatically determine how to validate that field. In the following example, we can see that a field of type non_neg_integer() will not accept a value &lt; 0: defmodule User do use Breakfast cereal do field :name, String.t() field :age, non_neg_integer() end end iex&gt; data = %{ ...&gt; &quot;name&quot; =&gt; &quot;Sean&quot;, ...&gt; &quot;age&quot; =&gt; -5 ...&gt; } ...&gt; Breakfast.decode(User, data) %Breakfast.Yogurt{ errors: [age: &quot;expected a non_neg_integer, got: -5&quot;], params: %{&quot;age&quot; =&gt; -5, &quot;name&quot; =&gt; &quot;Sean&quot;}, struct: %User{age: nil, name: &quot;Sean&quot;} } Breakfast can even handle more complex types, such as unions: defmodule Request do use Breakfast cereal do field :payload, map() field :status, :pending | :success | :failed end end iex&gt; data = %{ ...&gt; &quot;payload&quot; =&gt; %{&quot;some&quot; =&gt; &quot;data&quot;}, ...&gt; &quot;status&quot; =&gt; :success ...&gt; } ...&gt; Breakfast.decode(Request, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;payload&quot; =&gt; %{&quot;some&quot; =&gt; &quot;data&quot;}, &quot;status&quot; =&gt; :success}, struct: %Request{payload: %{&quot;some&quot; =&gt; &quot;data&quot;}, status: :success} } iex&gt; data = %{ ...&gt; &quot;payload&quot; =&gt; %{&quot;some&quot; =&gt; &quot;data&quot;}, ...&gt; &quot;status&quot; =&gt; :waiting ...&gt; } ...&gt; Breakfast.decode(Request, data) %Breakfast.Yogurt{ errors: [status: &quot;expected one of :pending | :success | :failed, got: :waiting&quot;], params: %{&quot;payload&quot; =&gt; %{&quot;some&quot; =&gt; &quot;data&quot;}, &quot;status&quot; =&gt; :waiting}, struct: %Request{payload: %{&quot;some&quot; =&gt; &quot;data&quot;}, status: nil} } Checkout the types docs for more on what types Breakfast supports."},{"ref":"readme.html#using-the-result","title":"README - Using the Result","type":"extras","doc":"You might be asking, what&#39;s this %Yogurt{} thing? A %Yogurt{} represents the result of a decoding. It contains three pieces of data: params: The original input params that you asked Breakfast to decode errors: A list of human-readable string errors that were accumulated when trying to decode the params struct: The decoded data that&#39;s been casted to the well-defined struct In your day-to-day programming, you can pattern match on a %Yogurt{} for control-flow, where an empty :errors list indicates that the decoding was successful: defmodule MathRequest do use Breakfast cereal do field :lhs, number() field :rhs, number() field :operation, :+ | :- | :* | :/, cast: :existing_atom_from_string end def existing_atom_from_string(value) do {:ok, String.to_existing_atom(value)} rescue _ in ArgumentError -&gt; :error end end iex&gt; request = %{&quot;lhs&quot; =&gt; 5.0, &quot;rhs&quot; =&gt; 2, &quot;operation&quot; =&gt; &quot;/&quot;} ...&gt; case Breakfast.decode(MathRequest, request) do ...&gt; %Breakfast.Yogurt{errors: [], struct: result} -&gt; {:ok, result} ...&gt; %Breakfast.Yogurt{errors: errors} -&gt; {:error, errors} ...&gt; end {:ok, %MathRequest{lhs: 5.0, rhs: 2, operation: :/}} iex&gt; request = %{&quot;lhs&quot; =&gt; 5.0, &quot;rhs&quot; =&gt; 2, &quot;operation&quot; =&gt; &quot;%&quot;} ...&gt; case Breakfast.decode(MathRequest, request) do ...&gt; %Breakfast.Yogurt{errors: [], struct: result} -&gt; {:ok, result} ...&gt; %Breakfast.Yogurt{errors: errors} -&gt; {:error, errors} ...&gt; end {:error, [operation: &quot;expected one of :+ | :- | :* | :/, got: :%&quot;]} What about :ok | :error tuples? We decided to not use :ok | :error tuples as the return type for the following reasons: We wanted to have a consistent type for the return value (it&#39;s always a Yogurt.t(), no matter what) There&#39;s a lot of context to return that you may or may not want to use (i.e. errors, input params, etc) You can still pattern match on any case that you care about handling in your code"},{"ref":"readme.html#custom-configuration","title":"README - Custom Configuration","type":"extras","doc":"When Breakfast is decoding data, it runs through the same 3 steps for each field: fetch: Retrieve the field value from the data (i.e. Map.fetch/2) cast: Map the field value from one value to another, if necessary (i.e. Integer.parse/1) validate: Check to see if the field value is valid (i.e. is_binary/1) Out of the box, Breakfast will assume the following for each step: fetch: The raw data is in the form of a string-keyed map, and the key for the field is the string version of the declared field name cast: No casting is necessary validate: Ensure that the value matches the field type However, each of these steps can be customized for any field: defmodule Settings do use Breakfast cereal do field(:name, String.t(), fetch: :fetch_name) field(:timeout, integer(), cast: :int_from_string) field(:volume, integer(), validate: :valid_volume) end def fetch_name(params, :name) do Map.fetch(params, &quot;SettingsName&quot;) end def int_from_string(value) do with true &lt;- is_binary(value), {int, &quot;&quot;} &lt;- Integer.parse(value) do {:ok, int} else _ -&gt; :error end end def valid_volume(volume) when volume in 0..100, do: [] def valid_volume(volume), do: [&quot;expected an integer in 0..100, got: \#{inspect(volume)}&quot;] end iex&gt; data = %{ ...&gt; &quot;SettingsName&quot; =&gt; &quot;Control Pannel&quot;, ...&gt; &quot;timeout&quot; =&gt; &quot;1500&quot;, ...&gt; &quot;volume&quot; =&gt; 8 ...&gt; } ...&gt; Breakfast.decode(Settings, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;SettingsName&quot; =&gt; &quot;Control Pannel&quot;, &quot;timeout&quot; =&gt; &quot;1500&quot;, &quot;volume&quot; =&gt; 8}, struct: %Settings{name: &quot;Control Pannel&quot;, timeout: 1500, volume: 8} } iex&gt; data = %{ ...&gt; &quot;name&quot; =&gt; &quot;Control Pannel&quot;, ...&gt; &quot;timeout&quot; =&gt; 1500, ...&gt; &quot;volume&quot; =&gt; -100 ...&gt; } ...&gt; Breakfast.decode(Settings, data) %Breakfast.Yogurt{ errors: [name: &quot;value not found&quot;, timeout: &quot;cast error&quot;, volume: &quot;expected an integer in 0..100, got: -100&quot;], params: %{&quot;name&quot; =&gt; &quot;Control Pannel&quot;, &quot;timeout&quot; =&gt; 1500, &quot;volume&quot; =&gt; -100}, struct: %Settings{name: nil, timeout: nil, volume: nil} } You can also set the default behaviour for any of these steps: defmodule RGBColor do use Breakfast cereal fetch: :fetch_upcase_key, cast: :int_from_string, validate: :valid_rgb_value do field :r, integer() field :g, integer() field :b, integer() end def fetch_upcase_key(params, field) do key = field |&gt; to_string() |&gt; String.upcase() Map.fetch(params, key) end def int_from_string(value) do with true &lt;- is_binary(value), {int, &quot;&quot;} &lt;- Integer.parse(value) do {:ok, int} else _ -&gt; :error end end def valid_rgb_value(value) when value in 0..255, do: [] def valid_rgb_value(value), do: [&quot;expected an integer between 0 and 255, got: \#{inspect(value)}&quot;] end iex&gt; data = %{&quot;R&quot; =&gt; &quot;10&quot;, &quot;G&quot; =&gt; &quot;20&quot;, &quot;B&quot; =&gt; &quot;30&quot;} ...&gt; Breakfast.decode(RGBColor, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;B&quot; =&gt; &quot;30&quot;, &quot;G&quot; =&gt; &quot;20&quot;, &quot;R&quot; =&gt; &quot;10&quot;}, struct: %RGBColor{b: 30, g: 20, r: 10} } iex&gt; data = %{&quot;r&quot; =&gt; &quot;10&quot;, &quot;G&quot; =&gt; &quot;Twenty&quot;, &quot;B&quot; =&gt; &quot;500&quot;} ...&gt; Breakfast.decode(RGBColor, data) %Breakfast.Yogurt{ errors: [r: &quot;value not found&quot;, g: &quot;cast error&quot;, b: &quot;expected an integer between 0 and 255, got: 500&quot;], params: %{&quot;B&quot; =&gt; &quot;500&quot;, &quot;G&quot; =&gt; &quot;Twenty&quot;, &quot;r&quot; =&gt; &quot;10&quot;}, struct: %RGBColor{b: nil, g: nil, r: nil} } Given this, Breakfast can actually decode any form of data, not just maps: defmodule SpreadsheetRow do use Breakfast @column_indices %{ name: 0, age: 1, email: 2 } cereal fetch: :fetch_at_list_index do field :name, String.t() field :age, non_neg_integer() field :email, String.t() end def fetch_at_list_index(data, field_name) do index = Map.fetch!(@column_indices, field_name) Enum.fetch(data, index) end end iex&gt; data = [&quot;Sully&quot;, 37, &quot;sully@aol.com&quot;] ...&gt; Breakfast.decode(SpreadsheetRow, data) %Breakfast.Yogurt{ errors: [], params: [&quot;Sully&quot;, 37, &quot;sully@aol.com&quot;], struct: %SpreadsheetRow{age: 37, email: &quot;sully@aol.com&quot;, name: &quot;Sully&quot;} }"},{"ref":"readme.html#required-fields-and-default-values","title":"README - Required Fields and Default Values","type":"extras","doc":"By default, Breakfast considers every field to be a required field. The only way to make a field &quot;optional&quot; is to provide a :default value for that field: defmodule Post do use Breakfast cereal do field :title, String.t() field :content, String.t() field :tags, [String.t()], default: [] end end iex&gt; data = %{ ...&gt; &quot;title&quot; =&gt; &quot;Cool Thing I Did&quot;, ...&gt; &quot;content&quot; =&gt; &quot;Thanks for reading!&quot;, ...&gt; } ...&gt; Breakfast.decode(Post, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;content&quot; =&gt; &quot;Thanks for reading!&quot;, &quot;title&quot; =&gt; &quot;Cool Thing I Did&quot;}, struct: %Post{ content: &quot;Thanks for reading!&quot;, tags: [], title: &quot;Cool Thing I Did&quot; } } iex&gt; data = %{ ...&gt; &quot;title&quot; =&gt; &quot;Cool Thing I Did&quot;, ...&gt; &quot;content&quot; =&gt; &quot;Thanks for reading!&quot;, ...&gt; &quot;tags&quot; =&gt; [&quot;blockchain&quot;, &quot;crypto&quot;] ...&gt; } ...&gt; Breakfast.decode(Post, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;content&quot; =&gt; &quot;Thanks for reading!&quot;, &quot;tags&quot; =&gt; [&quot;blockchain&quot;, &quot;crypto&quot;], &quot;title&quot; =&gt; &quot;Cool Thing I Did&quot;}, struct: %Post{ content: &quot;Thanks for reading!&quot;, tags: [&quot;blockchain&quot;, &quot;crypto&quot;], title: &quot;Cool Thing I Did&quot; } }"},{"ref":"readme.html#embedded-cereals","title":"README - Embedded Cereals","type":"extras","doc":"Breakfast allows you to use decoders within each other to describe the shape of nested data. Here, the Config decoder is used as the type for the User.config field: defmodule Player do defmodule Config do use Breakfast cereal do field :timezone, String.t() field :sleep_timeout, non_neg_integer() end end use Breakfast cereal do field :name, String.t() field :score, integer() field :config, {:cereal, Config} end end iex&gt; data = %{ ...&gt; &quot;name&quot; =&gt; &quot;Leo&quot;, ...&gt; &quot;score&quot; =&gt; 1600, ...&gt; &quot;config&quot; =&gt; %{ ...&gt; &quot;timezone&quot; =&gt; &quot;EST&quot;, ...&gt; &quot;sleep_timeout&quot; =&gt; 5000 ...&gt; } ...&gt; } ...&gt; Breakfast.decode(Player, data) %Breakfast.Yogurt{ errors: [], params: %{&quot;config&quot; =&gt; %{&quot;sleep_timeout&quot; =&gt; 5000, &quot;timezone&quot; =&gt; &quot;EST&quot;}, &quot;name&quot; =&gt; &quot;Leo&quot;, &quot;score&quot; =&gt; 1600}, struct: %Player{ name: &quot;Leo&quot;, score: 1600, config: %Player.Config{ sleep_timeout: 5000, timezone: &quot;EST&quot; } } } iex&gt; data = %{ ...&gt; &quot;name&quot; =&gt; &quot;Leo&quot;, ...&gt; &quot;score&quot; =&gt; 1600, ...&gt; &quot;config&quot; =&gt; %{ ...&gt; &quot;timezone&quot; =&gt; &quot;EST&quot;, ...&gt; &quot;sleep_timeout&quot; =&gt; -5000 ...&gt; } ...&gt; } ...&gt; Breakfast.decode(Player, data) %Breakfast.Yogurt{ errors: [config: [sleep_timeout: &quot;expected a non_neg_integer, got: -5000&quot;]], params: %{&quot;config&quot; =&gt; %{&quot;sleep_timeout&quot; =&gt; -5000, &quot;timezone&quot; =&gt; &quot;EST&quot;}, &quot;name&quot; =&gt; &quot;Leo&quot;, &quot;score&quot; =&gt; 1600}, struct: %Player{config: nil, name: &quot;Leo&quot;, score: 1600} }"},{"ref":"readme.html#current-state","title":"README - Current State","type":"extras","doc":"Breakfast 0.1 has been released! Further v0.1.x versions will include bug fixes, enhancements, etc. Breakfast 0.2 development will include all new major features and breaking changes. Check out out the roadmap to see what&#39;s coming/if you are looking to contribute!"},{"ref":"readme.html#contributing","title":"README - Contributing","type":"extras","doc":"Contributions are extremely welcome! This can take the form of pull requests and/or opening issues for bugs, feature requests, or general discussion. If you want to make some changes but aren&#39;t sure where to begin, I&#39;d be happy to help :). I&#39;d like to thank the following people who contributed to this project either via code and/or good ideas: @evuez @GeoffreyPS"},{"ref":"types.html","title":"Types","type":"extras","doc":"Types When defining a decoder in Breakfast, we use Elixir Typespecs to declare the types for each field. Breakfast aims to be type aware. By type aware we mean that Breakfast knows how to validate if a value is in fact a String.t(), non_neg_integer(), list(atom()), etc. The following are the types that Breakfast has formal support for: Basic Types: any() atom() map() struct() tuple() float() integer() neg_integer() non_neg_integer() pos_integer() list(type) nonempty_list(type) Literals: Atoms (:red, :blue) Integers (1, 1..10) Lists ([type], [], [...], [type, ...], [key: value_type]) Maps (%{}, %{key: value_type}, %{required(key_type) =&gt; value_type}, %{optional(key_type) =&gt; value_type}, %Struct{}, %Struct{key: value_type}) Tuples ({}, {type1, type2, ... typeN}) Built-in Types: term() binary() boolean() keyword() keyword(t) list() nonempty_list() mfa() module() number() Union Types (:red | :blue | :green, String.t() | atom()) Remote Types that resolve to a supported type Other Breakfast decoders ({:cereal, Config}) Note about remote types Elixir allows for remote type definitions (i.e. custom types). Elixir has many built-in remote types, like String.t(). Most of these remote types can be resolved to more basic Elixir types. In the case of String.t(), it resolves to binary() (at least in Elixir 1.9). If a built-in remote type can be resolved to a type Breakfast can understand, it can handle that remote type no problem. Breakfast can handle many built-in remote types, but there are a few more complex ones that it cannot. Improving this type support would be a great contribution! :-) Proceed with caution when using remote types, because they might resolve to something you don&#39;t expect. For example, as of Elixir 1.9, if you were to fully resolve the Enum.t() type to it&#39;s most terminal type, you get the following resolve chain: Enum.t() -&gt; Enumerable.t() -&gt; term() -&gt; any() In other words, setting a field&#39;s type to Enum.t() is no better than saying it&#39;s type is any(). Note about user-defined types Breakfast should also be able to handle most User-Defined Types, which are just remote types that you defined yourself. Breakfast can only understand a user-defined type if the module where that type was defined was compiled before Breakfast tries to understand it. If defined like so, Breakfast will fail to understand the type color() given the compile time constraint: defmodule Texture do use Breakfast @type color :: :red | :green | :blue cereal do field :color, color() end end The best way to ensure that Elixir will compile a typespec in time for Breakfast to start using it is to define the type in an external module, and then require that external module in your decoder module: # in lib/types.ex defmodule Types do @type color :: :red | :green | :blue end # in lib/texture.ex defmodule Texture do use Breakfast require Types cereal do field :color, Types.color() end end Note about cyclical types Elixir allows you to define cyclical types (types that refer to themselves recursively). Here is an example: defmodule Cycle do @type a :: b() @type b :: a() end If we tries to resolve the type Cycle.a(), we&#39;d see that it points to Cycle.b(), which points back to Cycle.a(), and so on. There&#39;s not much Breakfast can do in this case, and so using cyclical types is not allowed."}]